//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "EngCheckersBase.h"
#include <stdio.h>

#pragma package(smart_init)

#define ALL_MOVE_TYPES 1
#define CAPTURES_ONLY 0

//---------------------------------------------------------------------------

__fastcall EngCheckersBase::EngCheckersBase(TAlgorithmType nType)
  : GameEngine(nType)
{
  // make sure the priority is high enough to get run before
  // the UI task.  we will call Suspended from the main thread
  // regularly and if the priority is the same as the main thread
  // this would be highly inefficient

  MaxUniqueMoves = MAXUNIQUEMOVES;
}

// ---------------------------------------------------------------------------
// empty the board to its initial status

void EngCheckersBase::InitBoard(void)
{
  GameEngine::InitBoard();
}

// ---------------------------------------------------------------------------
// check if game is ended by checking if there are valid moves

bool EngCheckersBase::IsEnded(const int nColour,int *nScoreEnded)
{
  // see if a valid move can be found
  if (PlayerHasMove(nColour))
  {
    // there is at least one valid move for this player
    *nScoreEnded = SCORE_DRAW;
    return false;
  }

  // if no possible moves, game is lost
  // game is ended
  *nScoreEnded = -SCORE_ISWON;
  return true;
}

// ---------------------------------------------------------------------------
// execute move and change static board representation
// assume move nPos is valid for colour
// return true if a possible next capture can be done

bool EngCheckersBase::ApplyMove(const TSingleMove nPos,const int nColour,
    const int nMoveType)
{
  TSingleMove nMove;
  int nDiff;

  SetPieceAt(nPos.MoveTo,GetPieceAt(nPos.MoveFrom));
  SetPieceAt(nPos.MoveFrom,0);

  nDiff = abs(Board_Ver(nPos.MoveTo)-Board_Ver(nPos.MoveFrom));

  if (nDiff == 2)
  {
    // this is a capture move, make sure to remove the piece that was captured
    // the captured piece is one ahead of the vertical location and is in
    // the middle of the old and new horizontal position
    SetPieceAt( ((Board_Hor(nPos.MoveFrom)+Board_Hor(nPos.MoveTo))/2)
               +((Board_Ver(nPos.MoveFrom)+Board_Ver(nPos.MoveTo))/2)
                *BoardDimension,0);
    if (nMoveType == ALL_MOVE_TYPES)
    {
      // do iterate from here
      nMove = nPos;
      while (ApplyNextMove(&nMove,nColour));
    }

    return true;
  }

  // no capture move found
  // do promotion
  if ((nColour==1)&&(Board_Ver(nPos.MoveTo)==(BoardDimension-1)))
    SetPieceAt(nPos.MoveTo,5);
  if ((nColour==2)&&(Board_Ver(nPos.MoveTo)==0))
    SetPieceAt(nPos.MoveTo,6);

  return false;
}

// ---------------------------------------------------------------------------
// execute next move linked to the move we just did

bool EngCheckersBase::ApplyNextMove(TSingleMove *nMove,const int nColour)
{
  int nMoves,i;
  TMoveList aMoves;

  nMoves = GenerateMovesType(&aMoves,nColour,0,CAPTURES_ONLY);
  i=0;
  while (i<nMoves)
  {
    // loop through the next capture move
    if (aMoves.aMoves[i].MoveFrom == nMove->MoveTo)
    {
      *nMove = aMoves.aMoves[i];
      // if the starting position corresponds to the
      // previous ending position, do it
      // do not take into account branching of deep taking here
      // TODO: add second direction after capture
      // TODO: count number of pieces captured, needed for draughts
      return ApplyMove(aMoves.aMoves[i],nColour,0);
    }
    i++;
  }

  // no further capture moves found
  // do promotion if move ends at board end
  if ((nColour==1)&&(Board_Ver(nMove->MoveTo)==(BoardDimension-1)))
    SetPieceAt(nMove->MoveTo,5);
  if ((nColour==2)&&(Board_Ver(nMove->MoveTo)==0))
    SetPieceAt(nMove->MoveTo,6);

  return false;
}

// ---------------------------------------------------------------------------
// generate a list of valid moves
// start by allowing all types

int EngCheckersBase::GenerateMoves(TMoveList *aMoves,const int nColour)
{
  return GenerateMovesType(aMoves,nColour,0,ALL_MOVE_TYPES);
}

// ---------------------------------------------------------------------------
// generate a list of valid moves
// must be capture-moves or generic moves

int EngCheckersBase::GenerateMovesType(TMoveList *aMoves,
                                      const int nColour,
                                      const int nStartingPos,
                                      int nMode)
{
  register int i;

  aMoves->TotalMoves = 0;  // set number of foundmoves to zero

  if (nMode==ALL_MOVE_TYPES)
    // search all moves
    for (i=nStartingPos;i<Fields;i++)
      {
        if (AddValidFrom(i,nColour,aMoves,ALL_MOVE_TYPES))
        {
          // one capture move found, return only list of capture moves
          // this will result in a research through the complete tree
          // of all capture moves
          // in order to speed-up, we will search from the last position
          // searched, so only the capture move is researched
          GenerateMovesType(aMoves,nColour,0,CAPTURES_ONLY);
          break;
        }
      }
  else
    // search captures only
    for (i=nStartingPos;i<Fields;i++)
      AddValidFrom(i,nColour,aMoves,nMode);

  return aMoves->TotalMoves;
}

// ---------------------------------------------------------------------------
// add a move to the movelist, inline function for the AddValidFrom function
// only called from the above functions

void EngCheckersBase::AddValidMoveToList(TMoveList *aMoves,
          const int nFrom, const int nTo, const bool bCapture)
{
  int nMoves = aMoves->TotalMoves;

  aMoves->aMoves[nMoves].MoveFrom     = nFrom;
  aMoves->aMoves[nMoves].MoveTo       = nTo;
  aMoves->aMoves[nMoves].bCaptureMove = bCapture;
  aMoves->aMoves[nMoves].bValidMove   = true;
  aMoves->TotalMoves = nMoves + 1;
}

// ---------------------------------------------------------------------------
// check if we could move a stone from a certain position
// function used for the user interface

bool EngCheckersBase::IsValidFrom(const int nPos,const int nOwnColour)
{
  register int i;
  TMoveList aMoves;

  GenerateMoves(&aMoves,nOwnColour);

  // test if there is at least one move found
  if (aMoves.TotalMoves)
    // list of valid moves is built, check now if it appears in the list
    for (i=0;i<aMoves.TotalMoves;i++)
      if ( aMoves.aMoves[i].MoveFrom==nPos )
        // second check is not required for a number of games
        // function is used for UI interface only, so speed is not a concern
        return true;

  return false;
}

// ---------------------------------------------------------------------------
// return the move in a formatted way for display
// will be overwritten for most games, implemented here for
// simplifying the initial developement

int EngCheckersBase::MoveRepresentation(TSingleMove *nMove, char *sMove)
{
  return sprintf(sMove,"%02i%c%02i",
                 (int) nMove->MoveFrom,
                 nMove->bCaptureMove?'x':'-',
                 (int) nMove->MoveTo);
}

// ---------------------------------------------------------------------------
// convert this string back to a move

void EngCheckersBase::Str2Move(char *sMove,TSingleMove *nMove)
{
  nMove->MoveFrom = (sMove[0]-'0')*10 + (sMove[1]-'0');
  nMove->MoveTo   = (sMove[3]-'0')*10 + (sMove[4]-'0');
  nMove->bCaptureMove = (sMove[2]=='x');
}

// ---------------------------------------------------------------------------
// eof






