//---------------------------------------------------------------------------

#ifdef _DEBUG
#include <iostream>
#include <fstream>
#endif

#include <vcl.h>
#include <stdio.h>
#pragma hdrstop

#include "GameClasses.h"
#include "EngGame.h"
#include "AlgorithmBase.h"

#pragma package(smart_init)

//---------------------------------------------------------------------------
//   Important: Methods and properties of objects in VCL can only be
//   used in a method called using Synchronize, for example:
//
//      Synchronize(UpdateCaption);
//
//   where UpdateCaption could look like:
//
//      void __fastcall Unit1::UpdateCaption()
//      {
//        Form1->Caption = "Updated in a thread";
//      }
//---------------------------------------------------------------------------

__fastcall GameEngine::GameEngine(TAlgorithmType nType)
  : TThread(true)
  // CreateSuspended := true
{
  // make sure the priority is high enough to get run before
  // the UI task.  we will call Suspended from the main thread
  // regularly and if the priority is the same as the main thread
  // this would be highly inefficient

  //Priority = tpHighest;
  InitBoard();
  SetAlgorithm(nType);

  Depth = 2;
}

//---------------------------------------------------------------------------

__fastcall GameEngine::~GameEngine()
{
  delete Search;
}

//---------------------------------------------------------------------------
// set the algorithm to be used

void GameEngine::SetAlgorithm(TAlgorithmType nType)
{
  delete Search;  // delete it even if it was not assigned

  if (nType == ALG_RANDOM)
    Search = new AlgorithmRandom();
  else if (nType == ALG_STATIC_EVAL)
    Search = new AlgorithmStaticEval();
  else if (nType == ALG_BRUTE_FORCE)
    Search = new AlgorithmBruteForce();
  else if (nType == ALG_BRUTE_FWIN)
    Search = new AlgorithmBruteFWin();
  else if (nType == ALG_ALPHA_BETA)
    Search = new AlgorithmAlphaBeta();
  else if (nType == ALG_NEGASCOUT)
    Search = new AlgorithmNegascout();
}

//---------------------------------------------------------------------------
// calculate a move given the the position as filled in

void __fastcall GameEngine::Execute()
{
  while (!Terminated)
  {
    CalcMove();
    // this is indicating to the calling layer that the move has been
    // found
    Suspend();
  }
}

// ---------------------------------------------------------------------------
// set the colour for which the move has to be calculated
// on the execute method

void GameEngine::SetPlayer(const int nColour)
{
  nToMove = nColour;
}

// ---------------------------------------------------------------------------
// return horizontal coordinate of certain case

int GameEngine::Board_Hor(const int nPos)
{
  return nPos%BoardDimension;
}

// ---------------------------------------------------------------------------
// return vertical coordinate of certain case

int GameEngine::Board_Ver(const int nPos)
{
  return nPos/BoardDimension;
}

// ---------------------------------------------------------------------------
// check if a certain move is valid by comparing it to the list of
// valid moves

bool GameEngine::IsValidMove(TSingleMove &nPos,const int nColour)
{
  register int i;
  TMoveList aMoves;

  // test if there is at least one move found
  if (GenerateMoves(&aMoves,nColour))
    // list of valid moves is built, check now if it appears in the list
    for (i=0;i<aMoves.TotalMoves;i++)
      if (aMoves.aMoves[i]==nPos)
      {
        // function is used for UI interface only, so speed is not a concern
        // we copy down values like capture flags etc, they
        // are not filled in in the user interface,
        // so validating the move is a good opportunity to align
        // entered moves with calculated equivalent
        nPos = aMoves.aMoves[i];
        return true;
      }
      
  return false;
}

// ---------------------------------------------------------------------------
// log the board, in debug mode only

void GameEngine::LogBoard(void)
{
#ifdef _DEBUG
  register int i,j;

  clog << endl << "Board position" << endl << endl;

  for (i=BoardDimension-1;i>=0;i--)
  {
    for (j=0;j<BoardDimension;j++)
      clog << aBoard[j][i] << " ";
    clog << endl;
  }

  clog << endl;
#endif
}

// ---------------------------------------------------------------------------
// calculate a move with the given method

TSingleMove GameEngine::CalcMove(void)
{
#ifdef _DEBUG
  // in case of debugging, we log every entry in the function
  clog << "GameEngine Level " <<  Depth << " Player " <<
       ( nToMove == 1 ? "White" : "Black" ) <<
       " Algorithm " << Search->sName->c_str() << endl;
#endif

  nNodes = 0;

  LogBoard();

  nSelectedMove.init();
  Search->CalcBestMove(this,Depth,nToMove,&nSelectedMove);

#ifdef _DEBUG
  // in case of debugging, we log every entry in the function
  clog << "Nodes searched : " << nNodes << endl ;
#endif

  return nSelectedMove;
}

// ---------------------------------------------------------------------------
// called upon class creation, clear all variables

void GameEngine::InitBoard(void)
{
  int i;

  for (i=0;i<Fields;i++)
    SetPieceAt(i,-1);

  nCurrentMoveInHistory = -1;
  nSelectedMove.MoveTo = -1;

  while (ClearHistoryMove());  // clear all moves in the history list

}

// ---------------------------------------------------------------------------
// call repetively to fill the internal representation of the board
// based on external values

void GameEngine::SetPieceAt(const int nPos,const int nPiece)
{
  aBoard[Board_Hor(nPos)][Board_Ver(nPos)] = nPiece;
}

// ---------------------------------------------------------------------------
// get the stone at position

int GameEngine::GetPieceAt(const int nPos)
{
  return aBoard[Board_Hor(nPos)][Board_Ver(nPos)];
}

// ---------------------------------------------------------------------------
// check if game is ended
// in a standard game, it is ended only if there is no move
// possible for either of the players

bool GameEngine::IsEnded(const int nColour,int *nScoreEnded)
{
  // assume by default that score equals draw if the game is
  // ended
  *nScoreEnded = SCORE_DRAW;

  // see if a valid move can be found
  if (PlayerHasMove(1))
    // there is at least one valid move for this player
    return false;

  // check both parties
  if (PlayerHasMove(2))
    // there is at least one valid move for this player
    return false;

  // no valid move, so game is ended
  return true;
}

// ---------------------------------------------------------------------------
// check if a player has at least one valid move
// for that we will generate a movelist and see if it is bigger than
// zero

bool GameEngine::PlayerHasMove(const int nColour)
{
  TMoveList aMoves;

  return (GenerateMoves(&aMoves,nColour) != 0);
}

// ---------------------------------------------------------------------------

bool GameEngine::IsValidFrom(const int nPos,const int nColour)
{
  return true;
}

// ---------------------------------------------------------------------------
// calculate the value of the position
// basic calculation is one point per piece

signed long int GameEngine::EvaluatePos(const int nColour)
{
  register int i,j;

  signed long int liValue = 0;
  signed long int aValues[] = { 0 , 10 , -10 };
  // allow space for normal pieces and for queens

  for (i=0;i<BoardDimension;i++)
    for (j=0;j<BoardDimension;j++)
      liValue += aValues[aBoard[i][j]];

  return nColour==1?liValue:-liValue;
}

// ---------------------------------------------------------------------------

void __fastcall GameEngine::SetDepth(int value)
{
  if(FDepth != value) {
    FDepth = value;
  }
}

// ---------------------------------------------------------------------------

int __fastcall GameEngine::GetDepth()
{
  return FDepth;
}

// ---------------------------------------------------------------------------
// clear out the history table
// return false if no more moves are available

bool GameEngine::ClearHistoryMove(void)
{
  if (nCurrentMoveInHistory<0) return false;
  aHistory[nCurrentMoveInHistory].MoveFrom = -1;
  aHistory[nCurrentMoveInHistory].MoveTo   = -1;
  nCurrentMoveInHistory--;
  return true;
}

// ---------------------------------------------------------------------------
// get a move from the history list

void GameEngine::GetHistoryMove(int nIndex,TSingleMove *nMove)
{
  *nMove = aHistory[nIndex];
}

// ---------------------------------------------------------------------------
// get a move from the history list

int GameEngine::GetHistoryIndex(void)
{
  return nCurrentMoveInHistory;
}

// ---------------------------------------------------------------------------
// store a move in the history list

void GameEngine::SetHistoryMove(TSingleMove& nMove)
{
  nCurrentMoveInHistory++;
  aHistory[nCurrentMoveInHistory] = nMove;
}

// ---------------------------------------------------------------------------
// return the move in a formatted way for display
// will be overwritten for most games, implemented here for
// simplifying the initial developement

int GameEngine::MoveRepresentation(TSingleMove *nMove, char *sMove)
{
  return sprintf(sMove,"%i-%i",
                (int) nMove->MoveFrom,
                (int) nMove->MoveTo);
}

// ---------------------------------------------------------------------------
// if a game consist of multiplemoves like e.g. rochade in chess
// or multiple takes in checkers, we call this function until
// it returns false

bool GameEngine::ApplyNextMove(TSingleMove *nPos, const int nColour)
{
  return false;
}

// ---------------------------------------------------------------------------
// push the status of the board

void GameEngine::PushBoard(TBoard *aSavedBoard)
{
  register i,j;
  for (i=0;i<10;i++)
    for (j=0;j<10;j++)
      aSavedBoard->aBoard[i][j] = aBoard[i][j];
}

// ---------------------------------------------------------------------------
// restore the status of the board

void GameEngine::PopBoard(const TBoard aSavedBoard)
{
  register i,j;
  for (i=0;i<10;i++)
    for (j=0;j<10;j++)
      aBoard[i][j] = aSavedBoard.aBoard[i][j];
}

// ---------------------------------------------------------------------------
// eof


