//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "EngOthello.h"
#include <stdio.h>
#pragma package(smart_init)

//---------------------------------------------------------------------------

__fastcall OthelloEngine::OthelloEngine(TAlgorithmType nType)
  : GameEngine(nType)
{
  // make sure the priority is high enough to get run before
  // the UI task.  we will call Suspended from the main thread
  // regularly and if the priority is the same as the main thread
  // this would be highly inefficient

  Depth = 6;
  Fields = 64;
  MaxUniqueMoves = MAXUNIQUEMOVES;
}

// ---------------------------------------------------------------------------
// empty the board to its initial status

void OthelloEngine::InitBoard(void)
{
  GameEngine::InitBoard();

  BoardDimension = 8;

  SetPieceAt(27,1);
  SetPieceAt(28,2);
  SetPieceAt(35,2);
  SetPieceAt(36,1);
}

// ---------------------------------------------------------------------------
// test if game is ended, this is no valid moves
// for any of the players

bool OthelloEngine::IsEnded(const int nColour,int *nScoreEnded)
{
  // assume by default that score equals draw if the game is
  // ended
  *nScoreEnded = SCORE_DRAW;

  // see if a valid move can be found
  if (PlayerHasMove(1))
    // there is at least one valid move for this player
    return false;

  // check both parties
  if (PlayerHasMove(2))
    // there is at least one valid move for this player
    return false;

  // no valid move, so game is ended, now calculate
  // who has won
  *nScoreEnded = GameEngine::EvaluatePos(nColour);
  return true;
}

// ---------------------------------------------------------------------------
// check if a move is possible, if only a pass move
// is possible, this function returns zerp

bool OthelloEngine::PlayerHasMove(const int nColour)
{
  TMoveList aMoves;

  if (GenerateMoves(&aMoves,nColour) == 1)
    return (aMoves.aMoves[0].MoveTo != -1);
  else
    return true;
}

// ---------------------------------------------------------------------------
// generate a list of valid moves

int OthelloEngine::GenerateMoves(TMoveList *aMoves,const int nColour)
{
  register int i;
  int nFound = 0;

  // see if can take and add to list of valid moves
  for (i=0;i<Fields;i++)
    if (CanPlayAt(i,nColour))
    {
      aMoves->aMoves[nFound].bValidMove = true;
      aMoves->aMoves[nFound].MoveTo = i;
      aMoves->aMoves[nFound].MoveFrom = 0;
      nFound++;
    }

  if (nFound==0)
  {
    // no moves possible, store pass move
    aMoves->aMoves[nFound].bValidMove = true;
    aMoves->aMoves[nFound].MoveTo = -1;
    aMoves->aMoves[nFound].MoveFrom = 0;
    nFound++;
  }

  aMoves->TotalMoves = nFound;
  return nFound;
}

// ---------------------------------------------------------------------------
// execute move and change static board representation
// assume move nPos is valid for colour
// returning true means the move sequence is finished

bool OthelloEngine::ApplyMove(const TSingleMove nPos,
  const int nColour,const int nMoveType)
{
  int nXpos;
  int nYpos;

  if ((nPos.bValidMove)
        && (nPos.MoveTo != -1))
  {
    // make sure a pass move is not executed
    SetPieceAt(nPos.MoveTo,nColour);

    nXpos = Board_Hor(nPos.MoveTo);
    nYpos = Board_Ver(nPos.MoveTo);

    TurnInDir(nColour,nXpos,nYpos,-1,-1);
    TurnInDir(nColour,nXpos,nYpos,-1,0);
    TurnInDir(nColour,nXpos,nYpos,-1,1);
    TurnInDir(nColour,nXpos,nYpos,0,-1);
    TurnInDir(nColour,nXpos,nYpos,0,1);
    TurnInDir(nColour,nXpos,nYpos,1,-1);
    TurnInDir(nColour,nXpos,nYpos,1,0);
    TurnInDir(nColour,nXpos,nYpos,1,1);
  }

  return true;
}

// ---------------------------------------------------------------------------
// check if putting a stone somewhere is possible

bool OthelloEngine::CanPlayAt(const int nPos,const int nOwnColour)
{
  int nXpos;
  int nYpos;
  int nColour = GetPieceAt(nPos);

  if (nColour) return false;  // field occupied

  nXpos = Board_Hor(nPos);
  nYpos = Board_Ver(nPos);

  if (FindInDir(nOwnColour,nXpos,nYpos,-1,-1)) return true;
  if (FindInDir(nOwnColour,nXpos,nYpos,-1,0)) return true;
  if (FindInDir(nOwnColour,nXpos,nYpos,-1,1)) return true;
  if (FindInDir(nOwnColour,nXpos,nYpos,0,-1)) return true;
  if (FindInDir(nOwnColour,nXpos,nYpos,0,1)) return true;
  if (FindInDir(nOwnColour,nXpos,nYpos,1,-1)) return true;
  if (FindInDir(nOwnColour,nXpos,nYpos,1,0)) return true;
  if (FindInDir(nOwnColour,nXpos,nYpos,1,1)) return true;

  return false;
}

// ---------------------------------------------------------------------------
// see if you can capture a stone starting from position x till border

bool __fastcall OthelloEngine::FindInDir(const int nOwnColour,
                      const int nStartX,const int nStartY,
                      const int nDX,const int nDY)
{
  register int nXpos = nStartX+nDX;
  register int nYpos = nStartY+nDY;

  // test if we are out of bounds
  if (nXpos<0 || nXpos>7 || nYpos<0 || nYpos>7) return false;

  // test we are next to a foreign key, empty not allowed
  if (aBoard[nXpos][nYpos] != (3-nOwnColour)) return false;

  do
  {
    nXpos += nDX;
    nYpos += nDY;

    // test if we are out of bounds
    if (nXpos<0 || nXpos>7 || nYpos<0 || nYpos>7) return false;

    // if we find back an own stone, move would be valid
    if (aBoard[nXpos][nYpos] == nOwnColour) return true;

    // test we do not have an empty stone
  }  while (aBoard[nXpos][nYpos]);

  return false;
}

// ---------------------------------------------------------------------------
// capture all stones in a given direction
// as we know we can capture, we must not do any bound checks
// after the initial capture check

bool __fastcall OthelloEngine::TurnInDir(const int nOwnColour,
                      const int nStartX,const int nStartY,
                      const int nDX,const int nDY)
{
  if (!FindInDir(nOwnColour,nStartX,nStartY,nDX,nDY)) return false;
  // if there is no possible move in direction, no use to start turning
  // around

  register int nXpos = nStartX+nDX;
  register int nYpos = nStartY+nDY;

  do
  {
    // change colour of stone
    aBoard[nXpos][nYpos] = nOwnColour;

    nXpos += nDX;
    nYpos += nDY;

  } while (aBoard[nXpos][nYpos] != nOwnColour);

  return true;
}

// ---------------------------------------------------------------------------
// express a move as a string

int OthelloEngine::MoveRepresentation(TSingleMove *nMove, char *sMove)
{
  if ((nMove->bValidMove)
        && (nMove->MoveTo != -1))
  {
    return sprintf(sMove,"%c%i",
                'a'+(char)Board_Hor(nMove->MoveTo),
                Board_Ver(nMove->MoveTo)+1);
  }
  else
    return sprintf(sMove,"pass");
}

// ---------------------------------------------------------------------------
// convert this string back to a move

void OthelloEngine::Str2Move(char *sMove,TSingleMove *nMove)
{
  nMove->bValidMove = true;
  nMove->MoveFrom = 0;
  if (strcmp(sMove,"pass") == 0)
    nMove->MoveTo = -1;
  else
    nMove->MoveTo   = sMove[0]-'a' + (sMove[1]-'1')*BoardDimension;
}

// ---------------------------------------------------------------------------
// eof




