//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include <stdio.h>
#include <time.h>

#ifdef _DEBUG
#include <iostream>
#include <fstream>
#endif

#include "StrategicGames.h"
#include "GameClasses.h"
#include "EngineUI.h"
#include "MainUI.h"

#include "EngOthelloLev3.h"
#include "CheckersEngLev1.h"
#include "EngConnect4.h"
#include "EngDraughts.h"
#include "EngTicTacToe.h"

#define PGN_VERSION "[XPGN 0.2]"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TFrameEngineUI *FrameEngineUI;

extern GameUIData *mGameUI;
extern TGame      *mGameData;

//---------------------------------------------------------------------------
__fastcall TFrameEngineUI::TFrameEngineUI(TComponent* Owner)
  : TFrame(Owner)
{
  AlgorithmComboBox->ItemIndex = 4;
  ComboBoxLevel->Visible = (AlgorithmComboBox->ItemIndex > 1);
  ComboBoxLevel->ItemIndex = 4;
  DetailsMemo->Visible = true;
}

//---------------------------------------------------------------------------
// delete all created engines upon termination

__fastcall TFrameEngineUI::~TFrameEngineUI()
{
  DumpEngine();
}

//---------------------------------------------------------------------------
// general dump engine function, used when switching or when
// destroying

void TFrameEngineUI::DumpEngine()
{
  if (oEngine)
  {
    delete oEngine;
    oEngine = 0L;
  }
}

//---------------------------------------------------------------------------
// create engine based on UI settings

void TFrameEngineUI::CreateEngine()
{
  char szLevel[128];
  TAlgorithmType nAlgorithm=(TAlgorithmType)AlgorithmComboBox->ItemIndex;

  if(mGameUI->mGameType==GAMETYPE_OTHELLO)
  {
    oEngine = new OthelloEngineLev3(nAlgorithm);
  }
  else if(mGameUI->mGameType==GAMETYPE_CONNECT4)
  {
    oEngine = new Connect4Eng(nAlgorithm);
  }
  else if(mGameUI->mGameType==GAMETYPE_TICTACTOE)
  {
    oEngine = new EngTicTacToe(nAlgorithm);
  }
  else if(mGameUI->mGameType==GAMETYPE_DRAUGHTS)
  {
    oEngine = new EngDraughts(nAlgorithm);
  }
  else
  {
    oEngine = new CheckersEngLev1(nAlgorithm);
    // support simple or complicated checkers
    oEngine->nVariant = mGameUI->mGameType==GAMETYPE_CHECKERS_SIMPLE ? 1 : 0;
  }

  if (ComboBoxLevel->Visible)
  {
    if (ComboBoxLevel->ItemIndex == -1) // make sure a level is selected
      ComboBoxLevel->ItemIndex = 1;
    oEngine->Depth = ComboBoxLevel->ItemIndex + 1; // and fill the depth
  }
}

//---------------------------------------------------------------------------
// return false if a new game can be started
// allows for construct if(ResetDisAllowed()) return

bool TFrameEngineUI::ResetDisAllowed()
{
  if (oEngine != NULL)
    if (mGameData->nGameStatus == GAMESTATUS_PLAYING)
      return (Application->MessageBox("Start new game",
                                      "Game in progress",
                                      MB_YESNO)==IDNO);
  return false;
}


//---------------------------------------------------------------------------
// start a new game

void _fastcall TFrameEngineUI::NewGame()
{
  DumpEngine();
  CreateEngine();

  while (oEngine->ClearHistoryMove()); // clear all moves in the history

  mGameData->init();

  LabelValue->Caption = "0";
  LabelTime->Caption = "";
  DetailsMemo->Lines->Clear();

  UpdateDisplay();
}

//---------------------------------------------------------------------------
// let the computer calculate the move

bool TFrameEngineUI::CalcMove()
{
  int nMoves;
  bool bResult = false;
  TMoveList aMoves;
  TDateTime tStart,tEnd;

  AnimateComputerCalc->Active = true;
  AnimateComputerCalc->Visible = true;
  Screen->Cursor = crHourGlass;
  Application->ProcessMessages();

  nMoves = oEngine->GenerateMoves(&aMoves,mGameData->nToMove);
  if (nMoves)
  {
    oEngine->SetPlayer(mGameData->nToMove);
    LabelValue->Caption = "(calculating)";
    LabelTime->Caption = "00:00:00";
    Application->ProcessMessages();

    tStart = Now();
    if (ExperimentalCheckBox->Checked)
    {
      ExperimentalSearch();
    }

    #ifdef _MULTI_THREADED_
    oEngine->Resume();
    // we wait until the thread finds its move
    while (!oEngine->Suspended) Sleep(1);
    #else
    oEngine->CalcMove();
    #endif

    tEnd = Now();

    PlayValidMove(oEngine->nSelectedMove);
    bResult = true;
    LabelTime->Caption = TimeToStr(tEnd-tStart);
  }

  SwitchUser();

  Screen->Cursor = crDefault;
  AnimateComputerCalc->Active = false;
  AnimateComputerCalc->Visible = false;

  return bResult;
}

//---------------------------------------------------------------------------
// do search along with the different algorithms, and log the results of
// the search on screen

void TFrameEngineUI::ExperimentalSearch(void)
{
  TDateTime tStart,tEnd;
  int i;
  char sMove[256];
  char sDetail[256];
  int nTypes;

  // save the current setting
  TAlgorithmType nAlgorithm=(TAlgorithmType)AlgorithmComboBox->ItemIndex;

  // set up the list of algorithms to try
  TAlgorithmType aEvals[] = {
             // ALG_RANDOM,
             // ALG_STATIC_EVAL,
             ALG_BRUTE_FORCE,
             ALG_BRUTE_FWIN,
             ALG_ALPHA_BETA,
             ALG_NEGASCOUT };

  DetailsMemo->Lines->Clear();
  Application->ProcessMessages();
  nTypes = sizeof(aEvals)/sizeof(aEvals[0]);
  for (i=0;i<nTypes;i++)
  {
    tStart = Now();
    oEngine->SetAlgorithm(aEvals[i]);
    DetailsMemo->Lines->Add(oEngine->Search->sName->c_str());
    oEngine->nSelectedMove.MoveFrom = -1;
    oEngine->nSelectedMove.MoveTo = -1;
    oEngine->nSelectedMove.nScore = -SCORE_INFINITE;

    #ifdef _MULTI_THREADED_
    #pragma warn Multithreaded
    oEngine->Resume();
    // we wait until the thread finds its move
    while (!oEngine->Suspended) Sleep(1);
    #else
    oEngine->CalcMove();
    #endif

    tEnd = Now();

    oEngine->MoveRepresentation(&(oEngine->nSelectedMove),sMove);
    sprintf(sDetail,"%s %d %s %ld",
                  sMove,
                  oEngine->nSelectedMove.nScore,
                  TimeToStr(tEnd-tStart),
                  oEngine->nNodes);
    DetailsMemo->Lines->Add(sDetail);
  }

  // reset the original setting
  oEngine->SetAlgorithm(nAlgorithm);
}

//---------------------------------------------------------------------------
// test if a move can be played

bool TFrameEngineUI::IsValidMove(TSingleMove& nMove)
{
  return oEngine->IsValidMove(nMove,mGameData->nToMove);
}

//---------------------------------------------------------------------------
// test if starting field can be used

bool TFrameEngineUI::IsValidFrom(int nPos,int nColour)
{
  return oEngine->IsValidFrom(nPos,nColour);
}

//---------------------------------------------------------------------------
// will play the move assuming it is valid
// it will switch the status from init to playing if needed
// it will do nothing if game is ended

void TFrameEngineUI::PlayValidMove(TSingleMove nMove)
{
  char sMove[256];

  bool bNextMove;    // is there a next move available in the chain

  if (mGameData->nGameStatus != GAMESTATUS_ENDED)
  {
    mGameData->nGameStatus = GAMESTATUS_PLAYING;
    bNextMove = oEngine->ApplyMove(nMove,mGameData->nToMove,0);
    oEngine->SetHistoryMove(nMove);

    MainForm->UpdateBoard();

    oEngine->MoveRepresentation(&(nMove),sMove);
    DetailsMemo->Lines->Add(sMove);

    while (bNextMove)
    {
      bNextMove = oEngine->ApplyNextMove(&nMove,mGameData->nToMove);
      MainForm->UpdateBoard();
    }
  }
}

//---------------------------------------------------------------------------
// display the board as it is stored in the engine

bool TFrameEngineUI::UpdateDisplay(void)
{
  int nValue;
  TMoveList aMoves;
  std::string sMsgBoxText;
  int nScore;
  int nMoves;

  // display pieces on the screen
  MainForm->UpdateDisplay();

  if (mGameData->nGameStatus != GAMESTATUS_ENDED)
  {
    if (oEngine->IsEnded(mGameData->nToMove,&nScore))
    {
      mGameData->nGameStatus = GAMESTATUS_ENDED;
      if (((nScore>0) && (mGameData->nToMove==1))
        || ((nScore<0) && (mGameData->nToMove==2)))
      {
        sMsgBoxText = mGameUI->sWhiteName;
        sMsgBoxText.append(" won");
      }
      else if (((nScore>0) && (mGameData->nToMove==2))
        || ((nScore<0) && (mGameData->nToMove==1)))
      {
        sMsgBoxText = mGameUI->sBlackName;
        sMsgBoxText.append(" won");
      }
      else
        sMsgBoxText = "draw";

      //LabelValue->Caption = sMsgBoxText;
      DetailsMemo->Lines->Add(sMsgBoxText.c_str());
      Application->MessageBox(sMsgBoxText.c_str(),"The game is ended",MB_OK);
    }
    else
    {
      nValue = oEngine->EvaluatePos(1);
      LabelValue->Caption = nValue;
      nMoves = oEngine->GenerateMoves(&aMoves,mGameData->nToMove);
      PassBitBtn->Visible = (nMoves == 0);
      if ((nMoves == 1)&(aMoves.aMoves[0].MoveTo == -1))
        PassBitBtn->Visible = true;
    }
  }

  Application->ProcessMessages();

  return false;
}

//---------------------------------------------------------------------------

void __fastcall TFrameEngineUI::ButtonPassClick(TObject *Sender)
{
  // passbutton only becomes possible after some moves
  // so we can assume that we are still playing after a
  // pass move
  // only go will halt after a double pass
  // todo: generate a simple move and play as a move
  // so it can be stored to disk and reloaded

  PassBitBtn->Visible = false;
  mGameData->nGameStatus = GAMESTATUS_PLAYING;
  SwitchUser();
}

//---------------------------------------------------------------------------
// function will switch the global variable and will also start computer
// thinking

bool TFrameEngineUI::SwitchUser()
{
  bool lReturn = false;

  // this call will show if the game was ended before a switch takes place
  UpdateDisplay();

  mGameData->nToMove = 3 - mGameData->nToMove;
  if (mGameData->nToMove==mGameData->nComputerColour)
  {
    // switch user again, calc move will call switchuser
    CalcMove();
    // and switch was succesful
    lReturn = true;
  }

  // will make sure the player to move becomes visible
  // otherwise, it could be in the if statement
  UpdateDisplay();
  return lReturn;
}

//---------------------------------------------------------------------------
// mapper for the engine function

int TFrameEngineUI::GetStoneAt(int nPos)
{
  return oEngine->GetPieceAt(nPos);
}

//---------------------------------------------------------------------------
// select a new level based on the combobox value

void __fastcall TFrameEngineUI::ComboBoxLevelChange(TObject *Sender)
{
  oEngine->Depth = ComboBoxLevel->ItemIndex + 1; // and fill the depth
}

//---------------------------------------------------------------------------
// the second combo box depends on the selection of the first

void __fastcall TFrameEngineUI::AlgorithmComboBoxChange(TObject *Sender)
{
  ComboBoxLevel->Visible = (AlgorithmComboBox->ItemIndex > 1);
}

//---------------------------------------------------------------------------

void __fastcall TFrameEngineUI::ExperimentalCheckBoxClick(TObject *Sender)
{
  //DetailsMemo->Visible = ExperimentalCheckBox->Checked;
}

//---------------------------------------------------------------------------
// save current game with given filename

void TFrameEngineUI::SaveGame(AnsiString sFileName)
{
  FILE *hSaveGame;
  time_t currenttime;
  struct tm *tblock;
  int i;
  int nMaxMove;
  TSingleMove nCurrentMove;
  char sTime[256];
  char sMove[256];

  #define LF "\r\n"

  currenttime = time(NULL);
  tblock = localtime(&currenttime);
  strftime(sTime,sizeof(sTime),"%d-%m-%Y",tblock);

  hSaveGame = fopen(sFileName.c_str(),"w+b");

  fputs(PGN_VERSION LF,hSaveGame);
  fputs("[Game \"",hSaveGame);
  fputs(mGameUI->sName,hSaveGame);
  fputs("\"]" LF "[Site \"http://users.pandora.be/ai\"]",hSaveGame);
  fprintf(hSaveGame,LF "[Date \"%s\"]" LF "[" , sTime);
  fputs(mGameUI->sWhiteName,hSaveGame);
  fputs(" \"Player\"]" LF "[",hSaveGame);
  fputs(mGameUI->sBlackName,hSaveGame);
  fputs(" \"Computer\"]" LF,hSaveGame);
  fputs("[Result \"0-0\"]" LF ,hSaveGame);

  nMaxMove = oEngine->GetHistoryIndex();
  for (i=0;i<=nMaxMove;i++,i++)
  {
    oEngine->GetHistoryMove(i,&nCurrentMove);
    oEngine->MoveRepresentation(&nCurrentMove,sMove);
    fprintf(hSaveGame,"%d. " LF "%s" LF,
        i/2 + 1,
        sMove);

    if ((i+1)<=nMaxMove)
    {
      oEngine->GetHistoryMove(i+1,&nCurrentMove);
      oEngine->MoveRepresentation(&nCurrentMove,sMove);
      fprintf(hSaveGame,"%s",
        sMove);
    }
    //if (i%6)
      fprintf(hSaveGame,LF);
  }
  fputs("[end]" LF,hSaveGame);

  fclose(hSaveGame);
}

//---------------------------------------------------------------------------
// open a game with given filename

void TFrameEngineUI::OpenGame(AnsiString sFileName)
{
  FILE *hOpenGame;
  char sMove[256];
  int nCurrentMode = 0;
  TSingleMove nMove;

  NewGame();

  hOpenGame = fopen(sFileName.c_str(),"rb");

  // we read the first line of the file to see if the version is correct
  if(fgets(sMove,256,hOpenGame)!=NULL)
  {
    #ifdef _DEBUG
    std::clog << sMove;
    #endif
    // this version of the software is only compatible with
    // its own version
    if (strncmp(sMove,PGN_VERSION,strlen(PGN_VERSION)) != 0)
    {
      // difference found, exit
      Application->MessageBox("File format not compatible",
          sFileName.c_str(),MB_OK);
      fclose(hOpenGame);
      return;
    }
  }

  while (fgets(sMove,256,hOpenGame)!=NULL)
  {
    #ifdef _DEBUG
    std::clog << sMove;
    #endif
    if (sMove[0] != '[')
    {
      // a no comment line
      if (nCurrentMode!=0)
      {
        // a move white or black
        oEngine->Str2Move(sMove,&nMove);
        PlayValidMove(nMove);
        mGameData->nToMove = 3-mGameData->nToMove;
        UpdateDisplay();
      }

      if (++nCurrentMode==3) nCurrentMode = 0;
    }
  }


  fclose(hOpenGame);
}

//---------------------------------------------------------------------------
// eof

