// Alpha-Beta algorithm file

#ifdef _DEBUG
#include <iostream>
#include <fstream>
#endif

#include "EngGame.h"
#include "AlgorithmBase.h"

//---------------------------------------------------------------------------

AlgorithmAlphaBeta::AlgorithmAlphaBeta(void)
{
  sName->assign("AlphaBeta");
}

// ---------------------------------------------------------------------------
// Get the best move in a certain position
// go down a level if at this level a best move cannot be found

int AlgorithmAlphaBeta::CalcBestMove(GameEngine *oEngine,
                const int nLevel,
                const int nToMove,
                TSingleMove *nBestMoveHere)
{
  int nActualMove;
  int nBestMoveDown;      // variable actually not used as it only needs
                          // to be reported to the upper level only
  TMoveList aMoves;
  signed int nScore;
  register int i;
  TBoard aSavedBoard;
  int nLowerBound = -SCORE_ISWON;
  int nUpperBound = SCORE_ISWON;
  int nMoves;

  oEngine->nNodes++;

  #ifdef _DEBUG
  // in case of debugging, we log every entry in the function
  std::string sLevelName = "AlphaBeta::Calc ";

  //for (i=6;i>nLevel;i--)
  //  sLevelName.insert(0," ");

  clog << sLevelName << "Level " <<  nLevel << " Player " <<
       ( nToMove == 1 ? "White" : "Black" ) << " " ;

  LogMoveList(oEngine);

  nScore = oEngine->EvaluatePos(nToMove);

  //clog << endl << "(" << oEngine->nNodes << ") score = "
  //  << nScore << endl;
  #endif

  nBestMoveHere->init();
  oEngine->PushBoard(&aSavedBoard);
  nMoves = oEngine->GenerateMoves(&aMoves,nToMove);

  if (nMoves)
  {
    for (i=0;i<nMoves;i++)
    {
      oEngine->ApplyMove(aMoves.aMoves[i],nToMove,1);
      #ifdef _DEBUG
        aMoveArray[nLevel] = aMoves.aMoves[i];
        clog << sLevelName << " evaluating : level " << nLevel
            << " : Bestscore = " << nBestMoveHere->nScore
            << " : moves : ";
        LogMoveList(oEngine);
      #endif

      // here recurse or exit
      if (oEngine->IsEnded(nToMove,&nScore))
      {
        #ifdef _DEBUG
          oEngine->LogBoard();
        #endif
      }
      else
      {
        // only go deeper if not ended
        if (nLevel>1)
        {
          nScore = AlphaBetaLeaf(oEngine,nLevel-1,3-nToMove,-nUpperBound,-nLowerBound);
          // adjust score for taking the move to the fastest win
          // first - adjusting towards zero
          if (nScore > 0) nScore --; else if (nScore < 0) nScore ++;
        }
        else
          nScore = oEngine->EvaluatePos(nToMove);
      }
      oEngine->PopBoard(aSavedBoard);

      #ifdef _DEBUG
        clog << sLevelName << " level " <<  nLevel << " : score "
              << nScore << " : " ;
        LogMoveList(oEngine);
        aMoveArray[nLevel].bValidMove = false;
      #endif

      if (nScore>=nBestMoveHere->nScore)
      {
        #ifdef _DEBUG
          clog << sLevelName << " level " <<  nLevel
              << " : better or equal score " << nScore
              << " than score " << nBestMoveHere->nScore << endl;
        #endif
        *nBestMoveHere = aMoves.aMoves[i];
        nBestMoveHere->nScore = nScore;
      }
    }
  }

  return -(nBestMoveHere->nScore);
}

// ---------------------------------------------------------------------------

int AlgorithmAlphaBeta::AlphaBetaLeaf(GameEngine *oEngine,
    const int nLevel,const int nToMove,int nLowerBound,int nUpperBound)
{
  int nMoves;
  TMoveList aMoves;
  register int i;
  TBoard aSavedBoard;
  int nScore;
  int nBestScore = -SCORE_INFINITE;

  oEngine->nNodes++;

  #ifdef _DEBUG
  std::string sLevelName = "AlphaBeta::Leaf ";

  //for (i=6;i>nLevel;i--)
  //  sLevelName.insert(0," ");

  // in case of debugging, we log every entry in the function
  clog << sLevelName << "Level " <<  nLevel << " Player " <<
       ( nToMove == 1 ? "White" : "Black" ) << " Alpha " <<
       nLowerBound << " Beta " << nUpperBound << " : ";

  LogMoveList(oEngine);
  #endif

  oEngine->PushBoard(&aSavedBoard);
  nMoves = oEngine->GenerateMoves(&aMoves,nToMove);

  if (nMoves)
  {
    for (i=0;i<nMoves;i++)
    {
      oEngine->ApplyMove(aMoves.aMoves[i],nToMove,1);
      #ifdef _DEBUG
        aMoveArray[nLevel] = aMoves.aMoves[i];
      #endif

      // here recurse or exit
      if (oEngine->IsEnded(nToMove,&nScore))
      {
        #ifdef _DEBUG
          oEngine->LogBoard();
        #endif
      }
      else
      {
        // only go deeper if not ended
        if (nLevel>1)
        {
          nScore = AlphaBetaLeaf(oEngine,nLevel-1,3-nToMove,-nUpperBound,-nLowerBound);
          // adjust score for taking the move to the fastest win
          // first - adjusting towards zero
          if (nScore > 0) nScore --; else if (nScore < 0) nScore ++;
        }
        else
          nScore = oEngine->EvaluatePos(nToMove);
      }
      oEngine->PopBoard(aSavedBoard);

      #ifdef _DEBUG
        clog << sLevelName << " level " <<  nLevel << " : score "
              << nScore << " : " ;
        LogMoveList(oEngine);
        aMoveArray[nLevel].bValidMove = false;
      #endif

      if (nScore >= nBestScore)
      {
        nBestScore = nScore;
        #ifdef _DEBUG
          clog << sLevelName << "new better score "
             << nScore << endl;
        #endif
        if (nScore >= nUpperBound)
        {
          #ifdef _DEBUG
            clog << sLevelName << "causing beta break " << endl;
          #endif
          break;
        }
        if (nScore > nLowerBound)
        {
          #ifdef _DEBUG
            clog << sLevelName << "causing alpha adjust " << endl;
          #endif
          nLowerBound = nScore;
        }
      }
    }
  }
  return -(nBestScore);
}

// ---------------------------------------------------------------------------
// eof
