// Negascout algorithm file

#ifdef _DEBUG
#include <iostream>
#include <fstream>
#endif

#include "EngGame.h"
#include "AlgorithmBase.h"

//---------------------------------------------------------------------------

AlgorithmNegascout::AlgorithmNegascout(void)
{
  sName->assign("Negascout");
}

// ---------------------------------------------------------------------------
// Get the best move in a certain position
// go down a level if at this level a best move cannot be found

int AlgorithmNegascout::CalcBestMove(GameEngine *oEngine,
                const int nLevel,
                const int nToMove,
                TSingleMove *nBestMoveHere)
{
  int nBestScore;
  int nLowerBound = -SCORE_ISWON;
  int nUpperBound = SCORE_ISWON;
  TSingleMove nBestMoveDown;

  oEngine->nNodes++;

  nBestScore = NegascoutRoot(oEngine,nLevel,nToMove,nLowerBound,nUpperBound,
                             &nBestMoveDown);

  *nBestMoveHere = nBestMoveDown;
  nBestMoveHere->nScore = nBestScore;

  return nBestMoveHere->nScore;
}

// ---------------------------------------------------------------------------

int AlgorithmNegascout::NegascoutRoot(GameEngine *oEngine,
                int nLevel,
                int nToMove,
                int nLowerBound,
                int nUpperBound,
                TSingleMove *nBestMoveHere)
{
  int nScore;
  int nInternalUpperBound;
  TBoard aSavedBoard;
  int i;
  int nMoves;
  TMoveList aMoves;
  TSingleMove nBestMoveDown;

  oEngine->nNodes++;

#ifdef _DEBUG
  // in case of debugging, we log every entry in the function
  std::string sLevelName = "NegaScout::Calc ";

  //for (i=6;i>nLevel;i--)
  //  sLevelName.insert(0," ");

  clog << sLevelName << "Level " <<  nLevel << " Player " <<
       ( nToMove == 1 ? "White" : "Black" ) << " " ;

  LogMoveList(oEngine);
#endif

  if (nLevel==0)
  {
    // nBestMoveHere is not set
    nBestMoveHere->nScore = oEngine->EvaluatePos(nToMove);
  }
  else
  {
    // nBestMoveHere is not reset to zero here, assume it is set at the higher level
    nBestMoveHere->nScore = - SCORE_INFINITE;
    nInternalUpperBound = nUpperBound;

    oEngine->PushBoard(&aSavedBoard);
    nMoves = oEngine->GenerateMoves(&aMoves,nToMove);

    for (i=0;i<nMoves;i++)
    {
      oEngine->ApplyMove(aMoves.aMoves[i],nToMove,1);

      #ifdef _DEBUG
        aMoveArray[nLevel] = aMoves.aMoves[i];
        clog << sLevelName << " evaluating : level " << nLevel
            << " : Bestscore = " << nBestMoveHere->nScore
            << " : moves : ";
        LogMoveList(oEngine);
      #endif

      // here recurse or exit
      if (oEngine->IsEnded(nToMove,&nScore))
      {
        #ifdef _DEBUG
          oEngine->LogBoard();
        #endif
      }
      else
      {
        // only go deeper if not ended
        if (nLevel>1)
        {
          nBestMoveDown.init();
          nScore = NegascoutRoot(oEngine,
                         nLevel-1,
                         3-nToMove,
                         -nInternalUpperBound,
                         -max(nLowerBound,nBestMoveHere->nScore),
                         &nBestMoveDown);
          // adjust score for taking the move to the fastest win
          // first - adjusting towards zero
          if (nScore > 0) nScore --; else if (nScore < 0) nScore ++;
        }
        else
          nScore = oEngine->EvaluatePos(nToMove);
      }

      #ifdef _DEBUG
        clog << sLevelName << " level " <<  nLevel << " : score "
              << nScore << " : " ;
        LogMoveList(oEngine);
      #endif

      if (nScore>=nBestMoveHere->nScore)
      {
        *nBestMoveHere = aMoves.aMoves[i];
        if ((nInternalUpperBound==nUpperBound)||(nLevel<=2))
        {
          nBestMoveHere->nScore = nScore;
        }
        else
        {
          // research same move
          nBestMoveDown.init();
          nBestMoveHere->nScore = NegascoutRoot(oEngine,
                                      nLevel-1,
                                      3-nToMove,
                                      -nUpperBound,
                                      -nScore,
                                      &nBestMoveDown);
        }
        if (nBestMoveHere->nScore>=nUpperBound)
          break;

        nInternalUpperBound = max(nLowerBound,nBestMoveHere->nScore)+1;
      }

      oEngine->PopBoard(aSavedBoard);
      #ifdef _DEBUG
        aMoveArray[nLevel].bValidMove = false;
      #endif
    }
  }

  return -(nBestMoveHere->nScore);
}

// ---------------------------------------------------------------------------
// eof
